@@ -125,145 +125,155 @@
 ##  values of genes: genes that are constant in their non-zero values will have
 ##  an SD of 0 and therefore scaling them will result in division by 0.
 
+.rowNzRanges_dgCMatrix <- function(X, verbose=FALSE) {
+    res <- .Call("row_rngs_nzrngs_RsparseMatrix_R", as(X, "RsparseMatrix"),
+                 verbose=verbose)
+    res
+}
+
+.rowNzRanges_SVT_SparseArray_byrow <- function(X, verbose=FALSE) {
+    res <- .Call("row_rngs_nzrngs_SVT_SparseMatrix_R", X, verbose=verbose)
+    res
+}
+
+.rowNzRanges_SVT_SparseArray_transpose_C <- function(X, verbose=FALSE) {
+    res <- .Call("col_rngs_nzrngs_SVT_SparseMatrix_R", t(X), verbose=verbose)
+    res
+}
+
+## after discussions at from https://github.com/Bioconductor/SparseArray/issues/22
+
+.rowNzRanges_SVT_SparseArray <- function(X, verbose=FALSE) {
+    res <- .Call("rowbycols_rngs_nzrngs_SVT_SparseMatrix_R", X, verbose=verbose)
+    res
+}
+
+#' @importFrom SparseArray NaArray
+.fast_replace_zeros_with_NAs <- function(x) {
+    stopifnot(is(x, "SparseArray"))
+    naa <- NaArray(dim=dim(x), type=type(x), dimnames=dimnames(x))
+    naa@NaSVT <- x@SVT ## ASSUMING x@SVT HAS NO NA VALUES!!
+    naa
+}
+
+#' @importFrom SparseArray NaArray nzwhich
+.safe_replace_zeros_with_NAs <- function(x) {
+    naa <- NaArray(dim=dim(x), type=type(x), dimnames=dimnames(x))
+    nzidx <- nzwhich(x)
+    naa[nzidx] <- x[nzidx]
+    naa
+}
+
+#' @importFrom MatrixGenerics rowMins rowMaxs
+.rowNzRanges_SVT_SparseArray_rowbycols_R <- function(X, anyna=FALSE, verbose=FALSE) {
+    naa <- NULL
+    if (anyna)
+        naa <- .safe_replace_zeros_with_NAs(X)
+    else
+        naa <- .fast_replace_zeros_with_NAs(X)  # only if 'X' is guaranteed to be NA-free!
+
+    ranges1 <- cbind(rowMins(X, na.rm=TRUE), rowMaxs(X, na.rm=TRUE))
+    ranges2 <- suppressWarnings(cbind(rowMins(naa, na.rm=TRUE), rowMaxs(naa, na.rm=TRUE)))
+    allzeros <- ranges1[ , 1L] == 0L & ranges1[ , 2L] == 0L
+    ranges2[allzeros] <- NA_integer_
+    cbind(ranges1, ranges2)
+}
+
+#' @importFrom S4Arrays DummyArrayGrid read_block
+.rowNzRanges <- function(X, anyna=FALSE, verbose=FALSE) {
+    res <- NULL
+    if (is.matrix(X))
+        res <- rowRanges(X, na.rm=TRUE)
+    else if (is(X, "dgCMatrix"))
+        res <- .rowNzRanges_dgCMatrix(X, verbose=verbose)
+    else if (is(X, "SVT_SparseArray"))
+        res <- .rowNzRanges_SVT_SparseArray(X, verbose=verbose)
+    else if (is(X, "DelayedArray")) {
+        grid <- DummyArrayGrid(dim(X))
+        block <- read_block(X, grid[[1L]])
+        if (is_sparse(block)) ## input HDF5 may be sparse or not
+            res <- .rowNzRanges_SVT_SparseArray(block, verbose=verbose)
+        else
+            res <- rowRanges(X)
+    } else
+        cli_abort(c("x"=sprintf(".rowNzRanges: input object class %s not handled yet",
+                                class(X))))
+    res
+}
+
+
+
 #' @importFrom S4Arrays is_sparse
 #' @importFrom sparseMatrixStats rowRanges
 #' @importFrom DelayedArray blockApply setAutoBPPARAM rowRanges
 #' @importFrom cli cli_alert_warning cli_abort cli_alert_info
 #' @importFrom cli cli_progress_bar cli_progress_done
 #' @importFrom BiocParallel SerialParam bpnworkers bpiterate bpprogressbar
-.filterGenes <- function(expr, removeConstant=TRUE, removeNzConstant=TRUE,
-                         verbose=TRUE, BPPARAM=NULL) {
-    geneRanges <- NULL
-    ## open parallelism only if filtering have to be done for
-    ## more than 100 genes on more than 100 samples
-    if (!is.null(BPPARAM) && bpnworkers(BPPARAM) > 1 &&
-        nrow(expr) > 100 && ncol(expr) > 100) {
+.filterGenes <- function(expr, anyna=FALSE, removeConstant=TRUE,
+                         removeNzConstant=TRUE, verbose=TRUE, BPPARAM=NULL,
+                         maxmem=Inf) {
+    rowrngs <- NULL
 
-        if (is(expr, "DelayedArray")) {     ## DelayedArray::rowRanges() takes care of
-                                            ## grid layout with respect to parallelization
-            if (verbose) {
-                msg <- sprintf("Searching for genes/features with constant values using %d cores",
-                               as.integer(bpnworkers(BPPARAM)))
-                cli_alert_info(msg)
-                bpprogressbar(BPPARAM) <- TRUE
-            }
-            setAutoBPPARAM(BPPARAM=BPPARAM)
-            geneRanges <- rowRanges(expr)
-            setAutoBPPARAM(NULL)
-
-        } else {
-
-            n_chunks <- bpnworkers(BPPARAM)
-            if (verbose) {
-                msg <- sprintf("Searching for genes/features with constant values using %d cores",
-                               as.integer(bpnworkers(BPPARAM)))
-                cli_alert_info(msg)
-                idpb <- cli_progress_bar("Computing row ranges", total=n_chunks)
-            }
-            bpprogressbar(BPPARAM) <- FALSE
-            geneRanges <- bpiterate(.row_iter(expr, idpb, n_chunks),
-                                    rowRanges, na.rm=TRUE, useNames=FALSE,
-                                    REDUCE=rbind, reduce.in.order=TRUE, BPPARAM=BPPARAM)
-
-            if (verbose)
-                cli_progress_done(idpb)
-        }
-    } else {
-        BPPARAM <- NULL ## disable any other BPPARAM configuration
-        if (verbose)
-            cli_alert_info("Searching for genes/features with constant values")
-        geneRanges <- rowRanges(expr, na.rm=TRUE, useNames=FALSE)
+    if (verbose) {
+        if (!is_sparse(expr))
+            cli_alert_info("Searching for rows with constant values")
+        else
+            cli_alert_info("Searching for rows with constant (nonzero) values")
     }
 
-    constantGenes <- (geneRanges[, 1] == geneRanges[, 2])
+    ## returns a matrix with as many rows as 'expr' and 2 columns if 'expr'
+    ## is dense, and 4 columns if it is sparse, where the first two columns
+    ## correspond to the minimum and maximum values of each row, while the
+    ## third and fourth columns, if they exist, they correspond to the
+    ## minimum and maximum nonzero values of each row, which will be NAs if
+    ## there are no nonzero values.
+    rowrngs <- .processMatrixRows(expr, .rowNzRanges, anyna=anyna,
+                                  verbose=verbose, BPPARAM=BPPARAM, maxmem=maxmem)
 
-    if (any(constantGenes) || anyNA(constantGenes)) {
-        invalidGenes <- (constantGenes | is.na(constantGenes))
-        msg <- sprintf("%d genes/features with constant values throughout the samples",
-                       sum(invalidGenes))
+    constantRows <- (rowrngs[, 1] == rowrngs[, 2])
+    mask <- is.na(constantRows)
+    if (any(mask))
+        constantRows[mask] <- TRUE
+
+    constantNzRows <- invalidRows <- invalidNzRows <- rep(FALSE, nrow(expr))
+    if (ncol(rowrngs) > 2) { ## sparse input
+        constantNzRows <- (rowrngs[, 3] == rowrngs[, 4])
+        mask <- is.na(constantNzRows)
+        if (any(mask)) ## no nonzero values imply constant nonzero values
+            constantNzRows[mask] <- TRUE
+    }
+
+    if (verbose && any(constantRows)) {
+        msg <- sprintf("%d rows with constant values throughout the columns",
+                       sum(constantRows))
         cli_alert_warning(msg)
-        if (removeConstant) {
-            cli_alert_warning("Genes/features with constant values are discarded")
-            expr <- expr[!invalidGenes, ]
-        }
+        if (removeConstant)
+           cli_alert_warning("Rows with constant values are discarded")
     }
 
-    if (is_sparse(expr)) {
-        if (!is.null(BPPARAM) && bpnworkers(BPPARAM) > 1 &&
-            nrow(expr) > 100 && ncol(expr) > 100) {
+    nzmask <- constantNzRows & !constantRows
+    if (verbose && any(nzmask)) {
+        msg <- sprintf("%d rows with constant nonzero values throughout the samples",
+                       sum(nzmask))
+        cli_alert_warning(msg)
+        if (removeNzConstant)
+           cli_alert_warning("Rows with constant nonzero values are discarded")
+    }
 
-            if (verbose) {
-                msg <- sprintf("Searching for genes/features with constant nonzero values using %d cores",
-                               as.integer(bpnworkers(BPPARAM)))
-                cli_alert_info(msg)
-            }
-        } else if (verbose) {
-            BPPARAM <- NULL ## disable any other BPPARAM configuration
-            cli_alert_info("Searching for genes/features with constant nonzero values")
-        }
+    removemask <- rep(FALSE, nrow(expr))
+    if (removeConstant)
+         removemask <- constantRows
 
-        nzGeneRanges <- NULL
-        if (is(expr, "DelayedMatrix")) {
-            rowNonzeroRanges_byBlock <- function(block) {
-                t(sapply(t(block)@SVT, function(x) range(x[[1]])))
-            }
-            nworkers <- 1L
-            if (verbose && !is.null(BPPARAM) && bpnworkers(BPPARAM) > 1 &&
-                nrow(expr) > 100 && ncol(expr) > 100) {
-                bpprogressbar(BPPARAM) <- TRUE
-                nworkers <- bpnworkers(BPPARAM)
-            }
-            nzGeneRanges <- blockApply(expr, rowNonzeroRanges_byBlock,
-                                       grid=rowgridsize(expr, nworkers=nworkers),
-                                       as.sparse=is_sparse(expr), BPPARAM=BPPARAM)
-            nzGeneRanges <- do.call("rbind", nzGeneRanges)
-        } else if (is(expr, "dgCMatrix")) {
-            if (verbose && !is.null(BPPARAM) && bpnworkers(BPPARAM) > 1 &&
-                nrow(expr) > 100 && ncol(expr) > 100) {
-                bpprogressbar(BPPARAM) <- TRUE
-                nzGeneRanges <- do.call("rbind", bplapply(.sparse2columnList(t(expr)),
-                                                          FUN=range, BPPARAM=BPPARAM))
-            } else
-                nzGeneRanges <- do.call("rbind", lapply(.sparse2columnList(t(expr)),
-                                                        FUN=range))
-        } else if (is(expr, "SVT_SparseArray")) {
-            if (verbose && !is.null(BPPARAM) && bpnworkers(BPPARAM) > 1 &&
-                nrow(expr) > 100 && ncol(expr) > 100) {
-                bpprogressbar(BPPARAM) <- TRUE
-                nzGeneRanges <- do.call("rbind", bplapply(t(expr)@SVT,
-                                                          FUN=function(x) range[[1]],
-                                                          BPPARAM=BPPARAM))
-            } else
-                nzGeneRanges <- do.call("rbind", lapply(t(expr)@SVT,
-                                                        FUN=function(x) range(x[[1]])))
-        } else
-            cli_abort("x"="Uknown sparse matrix class")
+    if (removeNzConstant && any(nzmask))
+         removemask <- removemask | nzmask
 
-        stopifnot(is.matrix(nzGeneRanges)) ## QC
-        stopifnot(nrow(nzGeneRanges) == nrow(expr)) ## QC
+    if (any(removemask)) {
+        if (nrow(expr) - sum(removemask) < 2)
+            cli_abort(c("x"="Less than two rows left in the input assay object"))
 
-        ## nzGeneRanges <- NULL
-        ## if (bpnworkers(BPPARAM) > 1 && nrow(expr) > 100 && ncol(expr) > 100) {
-        ##   nzGeneRanges <- do.call("cbind", bplapply(nzGeneList, FUN=range, BPPARAM=BPPARAM))
-        ## } else
-        ##   nzGeneRanges <- vapply(nzGeneList, FUN=range, FUN.VALUE=double(2))
-        constantNzGenes <- (nzGeneRanges[, 1] == nzGeneRanges[, 2])
-
-        if (any(constantNzGenes) || anyNA(constantNzGenes)) {
-            invalidNzGenes <- (constantNzGenes | is.na(constantNzGenes))
-            msg <- sprintf("%d genes/features with constant nonzero values throughout the samples",
-                           sum(invalidNzGenes))
-            cli_alert_warning(msg)
-            if (removeNzConstant) {
-                cli_alert_warning("Genes/features with constant nonzero values are discarded")
-                expr <- expr[!invalidNzGenes, ]
-            }
-        }
+        expr <- expr[!removemask, ]
     }
 
-    if (nrow(expr) < 2)
-        cli_abort(c("x"="Less than two genes in the input assay object"))
-    
     return(expr)
 }
 
@@ -374,18 +384,17 @@
     
     ## filter genes according to various criteria,
     ## e.g., constant expression
-    filteredDataMatrix <- .filterGenes(dataMatrix,
-                                       removeConstant=removeConstant,
-                                       removeNzConstant=removeNzConstant,
-                                       verbose,
-                                       BPPARAM=BPPARAM)
+    filtDataMatrix <- .filterGenes(dataMatrix, anyna=anyNA(param),
+                                   removeConstant=removeConstant,
+                                   removeNzConstant=removeNzConstant,
+                                   verbose, BPPARAM=BPPARAM)
 
-    filteredMappedGeneSets <- .filterAndMapGeneSets(param=param,
-                                                    filteredDataMatrix=filteredDataMatrix,
-                                                    verbose=verbose)
+    filtMappedGeneSets <- .filterAndMapGeneSets(param=param,
+                                                filteredDataMatrix=filtDataMatrix,
+                                                verbose=verbose)
 
-    return(list(filteredDataMatrix=filteredDataMatrix,
-                filteredMappedGeneSets=filteredMappedGeneSets))
+    return(list(filteredDataMatrix=filtDataMatrix,
+                filteredMappedGeneSets=filtMappedGeneSets))
 }
 
 
@@ -461,6 +470,54 @@
     list(any_na=any_na, didCheckNA=didCheckNA)
 }
 
+## adapted from .define_multiworker_grid() in beachmat/R/colBlockApply.R
+#' @importFrom DelayedArray rowAutoGrid colAutoGrid getAutoBlockLength type
+.rowgridsize <- function(X, nworkers=1, maxmem=Inf) {
+  typesze <- c("integer"=4, "double"=8) ## 4 bytes for integers, 8 bytes for doubles
+  grid <- DummyArrayGrid(dim(X))
+  if (!is.infinite(maxmem) || nworkers > 1 || is(X, "DelayedMatrix")) {
+      ## initially maximum block length is the maximum of the default auto block
+      ## length and the maximum available memory divided by the size of stored number
+      ## if no finite maximum available memory is specified, then it becomes the
+      ## default auto block length
+      max.block.length <- getAutoBlockLength(type(X))
+      if (!is.infinite(maxmem))
+          max.block.length <- max(max.block.length, ceiling(maxmem / typesze[type(X)]))
+      ## assuming all workers share memory, the maximum block length has to reduce
+      ## by the number of workers to avoid exceeding the maximum available memory
+      ## and, in any case, it cannot exceed .Machine$integer.max
+      max.block.length <- min(.Machine$integer.max, max.block.length / nworkers)
+      expected.block.length <- max(1, ceiling(nrow(X) / nworkers) * as.numeric(ncol(X)))
+      block.length <- min(max.block.length, expected.block.length)
+      grid <- rowAutoGrid(X, block.length=block.length)
+  }
+  grid
+}
+
+## adapted from .define_multiworker_grid() in beachmat/R/colBlockApply.R
+#' @importFrom DelayedArray rowAutoGrid colAutoGrid getAutoBlockLength type
+.colgridsize <- function(X, nworkers=1, maxmem=Inf) {
+  typesze <- c("integer"=4, "double"=8) ## 4 bytes for integers, 8 bytes for doubles
+  grid <- DummyArrayGrid(dim(X))
+  if (!is.infinite(maxmem) || nworkers > 1 || is(X, "DelayedMatrix")) {
+      ## initially maximum block length is the maximum of the default auto block
+      ## length and the maximum available memory divided by the size of stored number
+      ## if no finite maximum available memory is specified, then it becomes the
+      ## default auto block length
+      max.block.length <- getAutoBlockLength(type(X))
+      if (!is.infinite(maxmem))
+          max.block.length <- max(max.block.length, ceiling(maxmem / typesze[type(X)]))
+      ## assuming all workers share memory, the maximum block length has to reduce
+      ## by the number of workers to avoid exceeding the maximum available memory
+      ## and, in any case, it cannot exceed .Machine$integer.max
+      max.block.length <- min(.Machine$integer.max, max.block.length / nworkers)
+      expected.block.length <- max(1, ceiling(nrow(X) / nworkers) * as.numeric(ncol(X)))
+      block.length <- min(max.block.length, expected.block.length)
+      grid <- colAutoGrid(X, block.length=block.length)
+  }
+  grid
+}
+
 #' @importClassesFrom IRanges IRanges
 #' @importFrom IRanges ranges
 .splitRowsInRanges <- function(grid) {
@@ -480,27 +537,32 @@
 ## through a BiocParallelParam object BPPARAM, when different from NULL, and
 ## reporting progress using the 'cli' package when possible
 
-#' @importFrom cli cli_abort cli_progress_bar
-#' @importFrom BiocParallel bplapply bpnworkers bpprogressbar
+#' @importFrom cli cli_abort cli_progress_bar cli_alert_warning
+#' @importFrom BiocParallel bplapply bpnworkers bpprogressbar bptry bpok
+#' @importFrom memuse howbig
 #' @importClassesFrom IRanges IRanges
 #' @importFrom IRanges start end width
-.processMatrixRows <- function(X, FUN, ..., progressmsg="Progress",
-                               verbose=TRUE, minparrows=100, minparcols=100,
-                               BPPARAM=NULL) {
+.processMatrixRows <- function(X, FUN, ..., verbose=TRUE,
+                               minparrows=100, minparcols=100,
+                               progressmsg="Progress", BPPARAM=NULL, maxmem=Inf) {
     stopifnot(length(dim(X)) == 2) ## QC
     FUN <- match.fun(FUN)
     nworkers <- 1L
     if (!is.null(BPPARAM) && nrow(X) > minparrows && ncol(X) > minparcols) {
         if (!is(BPPARAM, "BiocParallelParam"))
-            cli_abort("x"="'BPPARAM' must be a BiocParallelParam derivative")
+            cli_abort(c("x"="'BPPARAM' must be a BiocParallelParam derivative"))
         nworkers <- bpnworkers(BPPARAM)
     }
 
-    grid <- DummyArrayGrid(dim(X))
-    if (nworkers > 1 || is(X, "DelayedMatrix"))
-        grid <- rowgridsize(X, nworkers)
+    grid <- .rowgridsize(X, nworkers, maxmem)
     rir <- .splitRowsInRanges(grid)
-    if (length(rir) == 1)                     ## serial execution in one single call
+    if (length(rir) > 1 && verbose) {
+        typesze <- c("integer"=4, "double"=8) ## 4 bytes for integers, 8 bytes for doubles
+        sze <- howbig(as.numeric(width(rir[[1]])), as.numeric(ncol(X)),
+                      representation="dense", type=type(X))
+        cli_alert_info(sprintf("Splitting calculations in %d chunks of [%d, %d] and %s",
+                               length(rir), width(rir[[1]]), ncol(X), as.character(sze)))
+    } else if (length(rir) == 1)                     ## serial execution in one single call
         return(FUN(X, ..., verbose=verbose))
 
     FUN_WRAPPER <- function(rowsrng, verbose, idpbe, WRAPPED_FUN, ...) {
@@ -519,7 +581,8 @@
         env <- NULL
         if (verbose) {
             env <- new.env(parent=globalenv())
-            assign("idpb", cli_progress_bar(progressmsg, total=totalnrows), envir=env)
+            assign("idpb", cli_progress_bar(progressmsg, total=totalnrows),
+                   envir=env)
         }
         res <- lapply(rir, FUN=FUN_WRAPPER, verbose=verbose,
                       idpbe=env, WRAPPED_FUN=FUN, ...)
@@ -527,9 +590,24 @@
             cli_progress_done(get("idpb", envir=env))
     } else {                                  ## parallel execution in chunks
         if (verbose)
-            bpprogressbar(BPPARAM) <- TRUE    ## by now reporting progress wo/ cli
-        res <- bplapply(rir, FUN=FUN_WRAPPER, verbose=FALSE,
-                        idpbe=NULL, WRAPPED_FUN=FUN, ..., BPPARAM=BPPARAM)
+            bpprogressbar(BPPARAM) <- TRUE    ## reporting progress wo/ cli
+        bptry(res <- bplapply(rir, FUN=FUN_WRAPPER, verbose=FALSE,
+                              idpbe=NULL, WRAPPED_FUN=FUN, ...,
+                              BPPARAM=BPPARAM))
+        bpokmask <- bpok(res)
+        if (any(!bpokmask)) {
+            cli_alert_warning(sprintf("%d execution thread(s) give an error, reporting the first one"))
+            print(attr(res[[which(!bpokmask)]], "traceback"))
+            cli_alert_warning("Trying to execute again the failing thread(s)")
+            bptry(res <- bplapply(rir, FUN=FUN_WRAPPER, verbose=FALSE,
+                                  idpbe=NULL, WRAPPED_FUN=FUN, ...,
+                                  BPREDO=res, BPPARAM=BPPARAM))
+            if (any(!bpok(res))) {
+                cli_alert_warning(sprintf("%d execution thread(s) give an error, reporting the first one"))
+                print(attr(res[[which(!bpokmask)]], "traceback"))
+                cli_abort(c("x"="Cancelling execution"))
+            }
+        }
     }
     res <- do.call("rbind", res)
 
@@ -541,25 +619,31 @@
 ## reporting progress using the 'cli' package when possible
 
 #' @importFrom cli cli_abort
-#' @importFrom BiocParallel bplapply bpnworkers
+#' @importFrom BiocParallel bplapply bpnworkers bptry bpok
 #' @importClassesFrom IRanges IRanges
 #' @importFrom IRanges start end width
-.processMatrixCols <- function(X, FUN, ..., progressmsg="Progress",
-                               verbose=TRUE, minparrows=100, minparcols=100,
-                               BPPARAM=NULL) {
+.processMatrixCols <- function(X, FUN, ..., verbose=TRUE,
+                               minparrows=100, minparcols=100,
+                               progressmsg="Progress", BPPARAM=NULL, maxmem=Inf) {
     stopifnot(length(dim(X)) == 2) ## QC
     FUN <- match.fun(FUN)
     nworkers <- 1L
     if (!is.null(BPPARAM) && nrow(X) > minparrows && ncol(X) > minparcols) {
         if (!is(BPPARAM, "BiocParallelParam"))
-            cli_abort("x"="'BPPARAM' must be a BiocParallelParam derivative")
+            cli_abort(c("x"="'BPPARAM' must be a BiocParallelParam derivative"))
         nworkers <- bpnworkers(BPPARAM)
     }
-    grid <- DummyArrayGrid(dim(X))
-    if (nworkers > 1 || is(X, "DelayedMatrix"))
-        grid <- colgridsize(X, nworkers)
+
+    grid <- .colgridsize(X, nworkers, maxmem)
     cir <- .splitColsInRanges(grid)
-    if (length(cir) == 1)                     ## serial execution in one single call
+
+    if (length(cir) > 1 && verbose) {
+        typesze <- c("integer"=4, "double"=8) ## 4 bytes for integers, 8 bytes for doubles
+        sze <- howbig(as.numeric(nrow(X)), as.numeric(width(cir[[1]])),
+                      representation="dense", type=type(X))
+        cli_alert_info(sprintf("Splitting calculations in %d chunks of [%d, %d] and %s",
+                               length(cir), nrow(X), width(cir[[1]]), as.character(sze)))
+    } else if (length(cir) == 1)              ## serial execution in one single call
         return(FUN(X, ..., verbose=verbose))
 
     FUN_WRAPPER <- function(colsrng, verbose, idpbe, WRAPPED_FUN, ...) {
@@ -582,9 +666,24 @@
         cli_progress_done(get("idpb", envir=env))
     } else {                                  ## parallel execution in chunks
         if (verbose)
-            bpprogressbar(BPPARAM) <- TRUE    ## by now reporting progress wo/ cli
-        res <- bplapply(cir, FUN=FUN_WRAPPER, verbose=FALSE,
-                        idpbe=NULL, WRAPPED_FUN=FUN, ..., BPPARAM=BPPARAM)
+            bpprogressbar(BPPARAM) <- TRUE    ## reporting progress wo/ cli
+        bptry(res <- bplapply(cir, FUN=FUN_WRAPPER, verbose=FALSE,
+                              idpbe=NULL, WRAPPED_FUN=FUN, ...,
+                              BPPARAM=BPPARAM))
+        bpokmask <- bpok(res)
+        if (any(!bpokmask)) {
+            cli_alert_warning(sprintf("%d execution thread(s) give an error, reporting the first one"))
+            print(attr(res[[which(!bpokmask)]], "traceback"))
+            cli_alert_warning("Trying to execute again the failing thread(s)")
+            bptry(res <- bplapply(cir, FUN=FUN_WRAPPER, verbose=FALSE,
+                                  idpbe=NULL, WRAPPED_FUN=FUN, ...,
+                                  BPREDO=res, BPPARAM=BPPARAM))
+            if (any(!bpok(res))) {
+                cli_alert_warning(sprintf("%d execution thread(s) give an error, reporting the first one"))
+                print(attr(res[[which(!bpokmask)]], "traceback"))
+                cli_abort(c("x"="Cancelling execution"))
+            }
+        }
     }
     res <- do.call("cbind", res)
 
@@ -642,7 +741,7 @@
                 estimated_flag <- TRUE
             }
         } else
-            cli_abort("x"=sprintf("%s sparse matrix class cannot be handled", class(X)))
+            cli_abort(c("x"=sprintf("%s sparse matrix class cannot be handled", class(X))))
 
         if (verbose) {
             estmsg <- ""
@@ -657,6 +756,75 @@
     return(nzc)
 }
 
+#' @importFrom cli cli_abort
+.memtext2bytes <- function(x) {
+  if (is.numeric(x))
+      return(x)
+
+  x <- gsub(",", ".", x)
+  pat <- "(\\d*(.\\d+)*)(.*)"
+  num  <- as.numeric(sub(pat, "\\1", x))
+  unit <- sub(pat, "\\3", x)
+  unit[unit==""] <- "1"
+
+  fac <- c("1"=1, "K"=1024, "M"=1024^2, "G"=1024^3, "T"=1024^4)
+  if (!toupper(unit) %in% names(fac))
+      cli_abort(c("x"=sprintf("Unknown memory unit '%s', please use either K, M, G or T", unit)))
+
+  num * unname(fac[toupper(unit)])
+}
+
+#' @importFrom cli cli_abort cli_alert_info
+#' @importFrom memuse Sys.meminfo
+.check_maxmem <- function(param, x, verbose) {
+    if (length(x) > 1 || (!is.numeric(x) && !is.character(x)))
+        cli_abort(c("x"="'maxmem' should be a vector of length 1 of either a number in bytes or a character string"))
+
+    if (is.character(x) && x == "auto") {
+        totalram <- Sys.meminfo()$totalram
+        maxmem <- as.numeric(totalram * 0.9) ## auto takes 90% of RAM
+        X <- unwrapData(get_exprData(param), get_assay(param))
+        if (verbose && is(X, "DelayedArray") &&
+            gsva_global$show_start_and_end_messages)
+            cli_alert_info(sprintf("Maximum available main memory (90%%): %s",
+                                   as.character(totalram * 0.9)))
+    } else if (is.numeric(x))
+        maxmem <- x
+    maxmem <- .memtext2bytes(maxmem)
+    maxmem
+}
+
+#' @importFrom cli cli_abort cli_alert_info
+#' @importFrom S4Arrays is_sparse
+#' @importFrom memuse howbig
+.check_ondisk <- function(param, maxmem, verbose) {
+    ondisk <- get_ondisk(param)
+    if (ondisk == "auto") {
+        X <- unwrapData(get_exprData(param), get_assay(param))
+        tot <- as.numeric(nrow(X)) * as.numeric(ncol(X))
+        rep <- "dense"
+        spa <- 1
+        if (is_sparse(X)) {
+            rep <- "sparse"
+            spa <- nzcount(param) / tot
+        }
+        sze <- howbig(as.numeric(nrow(X)), as.numeric(ncol(X)),
+                      representation=rep, sparsity=spa, type=type(X))
+        ondisk <- "no"
+        if (as.numeric(sze) > maxmem) {
+            ondisk <- "yes"
+            if (is(X, "DelayedArray") && verbose)
+                cli_alert_info("On-disk input data does not fit in the maximum available main memory")
+        } else if (is(X, "DelayedArray") && verbose)
+            cli_alert_info("On-disk input data fits in the maximum available main memory")
+
+    } else if (ondisk != "yes" && ondisk != "no")
+        cli_abort(c("x"="'ondisk' should be either 'auto', 'yes' or 'no'"))
+
+    ondisk
+}
+
+
 ## transforms a dgCMatrix into a list of its
 ## non-zero values by MARGIN (1 for row, 2 for column)
 ##
